﻿// -------------------------------------------------------------------------------------------------
//          _____          _______                                   ____   _____
//      ,-:` \;',`'-,     |__  _(_)                                 |___ \ |  __ \
//    .'-;_,;  ':-;_,'.      | | _  _ __   _ __ ___    __ _  _ __     __) || |  | |
//   /;   '/    ,  _`.-\     | || || '_ \ | '_ ` _ \  / _` || '_ \   |__ < | |  | |
//  | '`. (`     /` ` \`|    | || || | | || | | | | || (_| || | | |  ___) || |__| |
//  |:.  `\`-.   \_   / |    |_||_||_| |_||_| |_| |_| \__,_||_| |_| |____/ |_____/
//  |     (   `,  .`\ ;'|
//   \     | .'     `-'/      Realtime terrain processing, rendering and analysis
//    `.   ;/        .'
//      `'-._____.-'`             http://www.tinman3d.com - info@tinman3d.com
// -------------------------------------------------------------------------------------------------
// FILE:       ModelScanner.cs
// -------------------------------------------------------------------------------------------------
// COPYRIGHT:  Matthias Englert, me@tinman3d.com
// -------------------------------------------------------------------------------------------------
// AUTHOR(s):  Matthias Englert, me@tinman3d.com
// -------------------------------------------------------------------------------------------------
// DISCLAIMER: This source  code is  provided "as  is" and  without warranties  as to performance or
//             merchantability. The  author and/or  distributors of  this source  code may have made
//             statements about this source code.  Any such statements do not  constitute warranties
//             and shall not be  relied on by  the user in deciding whether to use this source code.
//             This source code is provided without any expressed or implied warranties  whatsoever.
//             Because of the diversity of conditions and hardware under which this source code  may
//             be used,  no warranty  of fitness  for a  particular purpose  is offered. The user is
//             advised to test the source code thoroughly before relying on it. The user must assume
//             the entire risk of using the source code.
// -------------------------------------------------------------------------------------------------

using CodeX;
using Tinman.Core;
using Tinman.Core.IO;
using Tinman.Core.Math;
using Tinman.Core.System;
using Tinman.Core.Threading.Operations;
using Tinman.Core.Validating;
using Tinman.Engine.Models;
using Tinman.Terrain.Geometries;
using Tinman.Terrain.Heightmaps;

namespace Tinman.Demo.SIGGRAPH21
{
  /// <summary>
  ///   The <see cref="ModelScanner"/> class converts a 3D model to a heightmap.
  /// </summary>
  public sealed class ModelScanner : Disposable, IValidatable
  {
    #region Public / Attributes

    /// <summary>
    ///   Amount of blurring to apply to the generated displacement values before copying them to
    ///   the output dataset.
    /// </summary>
    /// <remarks>
    ///   Defaults to <c>5</c>.
    /// </remarks>
    /// <value>[>=1] The blur amount (see <see cref="CopySamplesOp.Blur"/>).</value>
    public int BlurDisplacement
    {
      get { return blurDisplacement; }
      set
      {
#if DEBUG
        if (value < 1)
          throw InvalidArgumentException.GreaterEqual("ModelScanner.BlurDisplacement", "value", 1);
#endif

        blurDisplacement = value;
      }
    }

    /// <summary>
    ///   Amount of blurring to apply to the generated elevation values before generating
    ///   displacement values and before copying them to the output dataset.
    /// </summary>
    /// <remarks>
    ///   Defaults to <c>50</c>.
    /// </remarks>
    /// <value>[>=1] The blur amount (see <see cref="CopySamplesOp.Blur"/>).</value>
    public int BlurElevation
    {
      get { return blurElevation; }
      set
      {
#if DEBUG
        if (value < 1)
          throw InvalidArgumentException.GreaterEqual("ModelScanner.BlurElevation", "value", 1);
#endif

        blurElevation = value;
      }
    }

    /// <summary>
    ///   Output path of the heightmap dataset that uses elevation and displacement to represent the
    ///   scanned surface of the 3D model.
    /// </summary>
    /// <remarks>
    ///   Defaults to <c>null</c>.
    /// </remarks>
    /// <value>The output path.</value>
    public Path Heightmap
    {
      get { return heightmap; }
      set { heightmap = value; }
    }

    /// <summary>
    ///   The 3D model to convert.
    /// </summary>
    /// <remarks>
    ///   Child models are not processed.
    /// </remarks>
    /// <value>The model to convert.</value>
    [OwnerValue]
    public IModelReader ModelReader
    {
      get { return modelReader; }
      set
      {
        if (modelReader != null)
        {
          modelReader.Dispose();
          modelReader = null;
        }

        modelReader = value;
      }
    }

    /// <summary>
    ///   The size of the generated heightmap datasets.
    /// </summary>
    /// <remarks>
    ///   Defaults to <c>257</c>.
    /// </remarks>
    /// <value>[pow2+1] The heightmap size.</value>
    public int Size
    {
      get { return size; }
      set
      {
#if DEBUG
        if (Precondition.NotPowerOfTwoPlusOne(value))
          throw InvalidArgumentException.PowerOfTwoPlusOne("ModelScanner.Size", "value");
#endif

        size = Maths.Clamp(value, HeightmapsUtil.MinSize, HeightmapsUtil.MaxSize);
      }
    }

    #endregion

    #region Public / Constructors

    /// <summary>
    ///   Creates a new instance of <see cref="ModelScanner"/>.
    /// </summary>
    public ModelScanner()
    {
      blurDisplacement = 5;
      blurElevation = 50;
      size = 257;
    }

    #endregion

    #region Public / Methods

    /// <summary>
    ///   Creates a background operation than scans the 3D model and generates a heightmap dataset.
    /// </summary>
    /// <remarks>
    ///   The following work is performed by the model scanner operation:
    ///   <list type="number">
    ///     <item>The 3D model is loaded from <see cref="ModelReader"/>, if necessary.</item>
    ///     <item>
    ///       The 3D model is analysed in order to determine a suitable base geometry for subsequent
    ///       scanning. Currently, the only supported base geometry is <see cref="Geometry.Sphere"/>.
    ///     </item>
    ///     <item>
    ///       The output dataset is created, using the given heightmap <see cref="Size"/>, the base
    ///       geometry and a geo-reference for a spherical planetary terrain.
    ///     </item>
    ///     <item>
    ///       For each sample of the output dataset, a ray is cast along the inverted up-direction
    ///       of the base geometry. Based on the intersection with the surface of the 3D model, a
    ///       terrain elevation value is computed and stored in a temporary elevation dataset.
    ///     </item>
    ///     <item>
    ///       The temporary elevation dataset is copied to the output dataset, using a copy
    ///       operation that blurs elevation values according to <see cref="BlurElevation"/>.
    ///     </item>
    ///     <item>
    ///       For each sample of the output dataset, a ray is cast along the normal vector of the
    ///       terrain surface defined by the blurred elevation values. Based on the intersection
    ///       with the surface of the 3D model, a terrain displacement value is computed and stored
    ///       in a temporary displacement dataset.
    ///     </item>
    ///     <item>
    ///       The temporary displacement dataset is copied to the output dataset, using a copy
    ///       operation that blurs displacement values according to <see cref="BlurDisplacement"/>.
    ///     </item>
    ///     <item>The temporary elevation and displacement datasets are deleted.</item>
    ///   </list>
    /// </remarks>
    /// <returns>[not-null] The background operation.</returns>
    /// <exception cref="ValidatingException">
    ///   If the model scanner settings are invalid.
    /// </exception>
    [OwnerReturn]
    public IOperation Scan()
    {
      this.ValidateThrowErrors("ModelScanner.Scan");

      return new ModelScanner_Operation(modelReader.Acquire(),
        heightmap,
        size,
        blurElevation,
        blurDisplacement);
    }

    public void Validate(Validator validator)
    {
#if DEBUG
      if (validator == null)
        throw InvalidArgumentException.NotNull("ModelScanner.Validate", "validator");
#endif

      if (heightmap == null)
        validator.MessageMember("Heightmap", ValidateMessage.MissingValue());

      if (modelReader == null)
        validator.MessageMember("ModelReader", ValidateMessage.MissingValue());
    }

    #endregion

    #region Protected / Methods

    protected override void DisposeResources()
    {
      if (modelReader != null)
      {
        modelReader.Dispose();
        modelReader = null;
      }

      heightmap = null;

      base.DisposeResources();
    }

    #endregion

    #region Private / Attributes

    int blurDisplacement;

    int blurElevation;

    Path heightmap;

    [Owner]
    IModelReader modelReader;

    int size;

    #endregion
  }
}