﻿// -------------------------------------------------------------------------------------------------
//          _____          _______                                   ____   _____
//      ,-:` \;',`'-,     |__  _(_)                                 |___ \ |  __ \
//    .'-;_,;  ':-;_,'.      | | _  _ __   _ __ ___    __ _  _ __     __) || |  | |
//   /;   '/    ,  _`.-\     | || || '_ \ | '_ ` _ \  / _` || '_ \   |__ < | |  | |
//  | '`. (`     /` ` \`|    | || || | | || | | | | || (_| || | | |  ___) || |__| |
//  |:.  `\`-.   \_   / |    |_||_||_| |_||_| |_| |_| \__,_||_| |_| |____/ |_____/
//  |     (   `,  .`\ ;'|
//   \     | .'     `-'/      Realtime terrain processing, rendering and analysis
//    `.   ;/        .'
//      `'-._____.-'`             http://www.tinman3d.com - info@tinman3d.com
// -------------------------------------------------------------------------------------------------
// FILE:       ModelScanner_Operation.cs
// -------------------------------------------------------------------------------------------------
// COPYRIGHT:  Matthias Englert, me@tinman3d.com
// -------------------------------------------------------------------------------------------------
// AUTHOR(s):  Matthias Englert, me@tinman3d.com
// -------------------------------------------------------------------------------------------------
// DISCLAIMER: This source  code is  provided "as  is" and  without warranties  as to performance or
//             merchantability. The  author and/or  distributors of  this source  code may have made
//             statements about this source code.  Any such statements do not  constitute warranties
//             and shall not be  relied on by  the user in deciding whether to use this source code.
//             This source code is provided without any expressed or implied warranties  whatsoever.
//             Because of the diversity of conditions and hardware under which this source code  may
//             be used,  no warranty  of fitness  for a  particular purpose  is offered. The user is
//             advised to test the source code thoroughly before relying on it. The user must assume
//             the entire risk of using the source code.
// -------------------------------------------------------------------------------------------------

using CodeX;
using Tinman.Core;
using Tinman.Core.Collections;
using Tinman.Core.Config;
using Tinman.Core.IO;
using Tinman.Core.IO.Files;
using Tinman.Core.Math;
using Tinman.Core.Math.Vectors;
using Tinman.Core.Threading.Operations;
using Tinman.Core.Threading.Pooling;
using Tinman.Engine.Models;
using Tinman.Terrain.Geometries;
using Tinman.Terrain.Georef;
using Tinman.Terrain.Heightmaps;
using Tinman.Terrain.Heightmaps.Formats;
using Tinman.Terrain.Imaging;
using Tinman.Terrain.Mapping;

namespace Tinman.Demo.SIGGRAPH21
{
  internal sealed class ModelScanner_Operation : Operation
  {
    #region Configuration

    public override ConfigValue ToConfig()
    {
      return ConfigValue.Invalid;
    }

    #endregion

    #region Public / Attributes

    public override string Name
    {
      get { return "Model Scanner"; }
    }

    #endregion

    #region Protected / Methods

    protected override void DisposeResources()
    {
      if (taskGroup != null)
      {
        taskGroup.Dispose();
        taskGroup = null;
      }

      if (modelReader != null)
      {
        modelReader.Dispose();
        modelReader = null;
      }

      heightmap = null;
      taskScanRegion = null;

      base.DisposeResources();
    }

    protected override void DoRun(IProgressMonitor progress)
    {
      IModel model;
      IModelCollider collider;
      Vec3D center;
      RangeD elevationRange;
      IArrayVector<Vec3D> vertices;
      double surface;
      int heightmapSize;
      VerticalRange verticalRange;
      Raster raster;
      IHeightmap heightmap; // [!]

      progress.TicksPending(4);

      // Load 3D model and get collider.

      model = modelReader.ReadModel();
      collider = model.Collider;

      // Use center-of-mass as the center of the scanned 'planetary' body, then compute the range
      // of elevation values, based on the vertices in the 3D model which are used to define
      // triangles.

      center = collider.Mass;

      elevationRange = RangeD.Inv;

      vertices = collider.Vertices();

      for (int i = 0; i < vertices.Count; i++)
        elevationRange = elevationRange.Grow(vertices.GetAt(i).Sub(center).Length);

      surface = (elevationRange.Start + elevationRange.End) / 2;

      heightmapSize = size;

      progress.TicksPending(12L * heightmapSize * heightmapSize);
      progress.TicksFinished(1);

      // Compute vertical range and geo-reference raster.

      verticalRange =
        new VerticalRange(elevationRange.Start - surface, elevationRange.End - surface);

      raster = new Raster(new CoordinateSystem(new GeodeticDatum(new Ellipsoid(surface, surface),
        new PrimeMeridian(0))));

      // Create output dataset.

      heightmap = Heightmap.Builder(heightmapSize)
        .Range(verticalRange)
        .Georef(raster)
        .Dataset(this.heightmap,
          HeightmapFormat.ForAll(
            HeightmapLayer.Coverage | HeightmapLayer.Elevation | HeightmapLayer.Displacement,
            PixelFormat.SixteenBits));

      try
      {
        IGeometry geometry;
        Path tempPath;
        IHeightmap temp; // [!]

        geometry = Geometry.ForGeoref(heightmap);

        tempPath = LocalFileSystem.FileTempPath("hgt");

        // Create temporary elevation dataset.

        temp = Heightmap.Builder(heightmap)
          .Dataset(tempPath,
            HeightmapFormat.ForAll(HeightmapLayer.Coverage | HeightmapLayer.Elevation,
              PixelFormat.SixteenBits));

        // Perform ray-casting to compute temporary elevation values.

        try
        {
          ITaskQueue<ModelScanner_Operation_Job, ModelScanner_Operation_Job> taskQueue;
          IOperation copySamples; // [!]

          taskQueue = taskScanRegion.Queue(false, false, 64);

          for (CubemapFace face = Cubemap.FirstFace; face <= Cubemap.LastFace; face++)
          {
            for (int y = 0; y < heightmapSize; y += BlockSize)
            {
              for (int x = 0; x < heightmapSize; x += BlockSize)
              {
                CubemapFaceRect region;
                ModelScanner_Operation_Job job;

                if (ShallCancel)
                {
                  taskGroup.Cancel(false);
                  return;
                }

                region = new CubemapFaceRect(x,
                  y,
                  Maths.Min(BlockSize, heightmapSize - x),
                  Maths.Min(BlockSize, heightmapSize - y),
                  face);

                job = new ModelScanner_Operation_Job(region,
                  temp,
                  geometry,
                  center,
                  verticalRange,
                  surface,
                  collider);

                while (taskQueue.IsReady)
                  progress.TicksFinished(taskQueue.Pop().Ticks);

                if (taskQueue.IsFull)
                  progress.TicksFinished(taskQueue.Pop().Ticks);

                taskQueue.Push(job);
              }
            }
          }

          while (!taskQueue.IsEmpty)
          {
            if (ShallCancel)
            {
              taskGroup.Cancel(false);
              return;
            }

            progress.TicksFinished(taskQueue.Pop().Ticks);
          }

          temp.Flush();

          progress.TicksFinished(1);

          // Blur temporary elevation values and copy to output dataset.

          copySamples = temp.Acquire()
            .CopySamples(this.heightmap,
              null,
              CopySamplesOp.Blur(blurElevation, HeightmapLayer.Elevation),
              PixelFormat.Default,
              false);

          try
          {
            copySamples.Run(progress);
          }
          finally
          {
            copySamples.Dispose();
          }
        }
        finally
        {
          temp.Dispose();
        }

        tempPath.FileDelete();

        // Create temporary displacement dataset.

        temp = Heightmap.Builder(heightmap)
          .Dataset(tempPath,
            HeightmapFormat.ForAll(HeightmapLayer.Coverage | HeightmapLayer.Displacement,
              PixelFormat.SixteenBits));

        // Perform ray-casting to compute temporary displacement values.

        try
        {
          ITaskQueue<ModelScanner_Operation_Job, ModelScanner_Operation_Job> taskQueue;
          IOperation copySamples; // [!]

          taskQueue = taskScanRegion.Queue(false, false, 64);

          for (CubemapFace face = Cubemap.FirstFace; face <= Cubemap.LastFace; face++)
          {
            for (int y = 0; y < heightmapSize; y += BlockSize)
            {
              for (int x = 0; x < heightmapSize; x += BlockSize)
              {
                CubemapFaceRect region;
                ModelScanner_Operation_Job job;

                if (ShallCancel)
                {
                  taskGroup.Cancel(false);
                  return;
                }

                region = new CubemapFaceRect(x,
                  y,
                  Maths.Min(BlockSize, heightmapSize - x),
                  Maths.Min(BlockSize, heightmapSize - y),
                  face);

                job = new ModelScanner_Operation_Job(region,
                  heightmap,
                  temp,
                  geometry,
                  center,
                  verticalRange,
                  collider);

                while (taskQueue.IsReady)
                  progress.TicksFinished(taskQueue.Pop().Ticks);

                if (taskQueue.IsFull)
                  progress.TicksFinished(taskQueue.Pop().Ticks);

                taskQueue.Push(job);
              }
            }
          }

          while (!taskQueue.IsEmpty)
          {
            if (ShallCancel)
            {
              taskGroup.Cancel(false);
              return;
            }

            progress.TicksFinished(taskQueue.Pop().Ticks);
          }

          temp.Flush();

          progress.TicksFinished(1);

          // Blur temporary displacement values and copy to output dataset.

          copySamples = temp.Acquire()
            .CopySamples(this.heightmap,
              null,
              CopySamplesOp.Blur(blurDisplacement, HeightmapLayer.Displacement),
              PixelFormat.Default,
              false);

          try
          {
            copySamples.Run(progress);
          }
          finally
          {
            copySamples.Dispose();
          }
        }
        finally
        {
          temp.Dispose();
        }

        tempPath.FileDelete();

        heightmap.Flush();

        progress.TicksFinished(1);
      }
      finally
      {
        heightmap.Dispose();
      }
    }

    #endregion

    #region Private / Constants

    const int BlockSize = 64;

    #endregion

    #region Private / Attributes

    readonly int blurDisplacement;

    readonly int blurElevation;

    Path heightmap;

    [Owner]
    IModelReader modelReader;

    readonly int size;

    [Owner]
    TaskGroup taskGroup;

    ITask<ModelScanner_Operation_Job, ModelScanner_Operation_Job> taskScanRegion;

    #endregion

    #region Private / Constructors

    internal ModelScanner_Operation([Owner] IModelReader modelReader, Path heightmap, int size,
      int blurElevation, int blurDisplacement)
    {
      this.modelReader = modelReader;
      this.heightmap = heightmap;
      this.size = size;
      this.blurElevation = blurElevation;
      this.blurDisplacement = blurDisplacement;

      taskGroup = new TaskGroup("ModelScanner");
      taskScanRegion =
        Task<ModelScanner_Operation_Job, ModelScanner_Operation_Job>.Wrap(ScanRegion,
          "ScanRegion",
          TaskFlags.None,
          taskGroup);
    }

    #endregion

    #region Private / Methods

    static ModelScanner_Operation_Job ScanRegion(ModelScanner_Operation_Job input)
    {
      input.Perform();

      return input;
    }

    #endregion
  }
}