﻿// -------------------------------------------------------------------------------------------------
//          _____          _______                                   ____   _____
//      ,-:` \;',`'-,     |__  _(_)                                 |___ \ |  __ \
//    .'-;_,;  ':-;_,'.      | | _  _ __   _ __ ___    __ _  _ __     __) || |  | |
//   /;   '/    ,  _`.-\     | || || '_ \ | '_ ` _ \  / _` || '_ \   |__ < | |  | |
//  | '`. (`     /` ` \`|    | || || | | || | | | | || (_| || | | |  ___) || |__| |
//  |:.  `\`-.   \_   / |    |_||_||_| |_||_| |_| |_| \__,_||_| |_| |____/ |_____/
//  |     (   `,  .`\ ;'|
//   \     | .'     `-'/      Realtime terrain processing, rendering and analysis
//    `.   ;/        .'
//      `'-._____.-'`             http://www.tinman3d.com - info@tinman3d.com
// -------------------------------------------------------------------------------------------------
// FILE:       ModelScanner_Operation_Job.cs
// -------------------------------------------------------------------------------------------------
// COPYRIGHT:  Matthias Englert, me@tinman3d.com
// -------------------------------------------------------------------------------------------------
// AUTHOR(s):  Matthias Englert, me@tinman3d.com
// -------------------------------------------------------------------------------------------------
// DISCLAIMER: This source  code is  provided "as  is" and  without warranties  as to performance or
//             merchantability. The  author and/or  distributors of  this source  code may have made
//             statements about this source code.  Any such statements do not  constitute warranties
//             and shall not be  relied on by  the user in deciding whether to use this source code.
//             This source code is provided without any expressed or implied warranties  whatsoever.
//             Because of the diversity of conditions and hardware under which this source code  may
//             be used,  no warranty  of fitness  for a  particular purpose  is offered. The user is
//             advised to test the source code thoroughly before relying on it. The user must assume
//             the entire risk of using the source code.
// -------------------------------------------------------------------------------------------------

using CodeX;
using Tinman.Core;
using Tinman.Core.Math;
using Tinman.Core.Math.Vectors;
using Tinman.Engine.Models;
using Tinman.Terrain.Buffers;
using Tinman.Terrain.Geometries;
using Tinman.Terrain.Heightmaps;
using Tinman.Terrain.Mapping;

namespace Tinman.Demo.SIGGRAPH21
{
  internal sealed class ModelScanner_Operation_Job
  {
    #region Public / Attributes

    public long Ticks
    {
      get { return (long) region.Width * region.Height; }
    }

    #endregion

    #region Private / Attributes

    readonly Vec3D center;

    readonly IModelCollider collider;

    readonly bool displacement;

    readonly IHeightmap elevation;

    readonly IGeometry geometry;

    readonly IHeightmap heightmap;

    readonly VerticalRange range;

    readonly CubemapFaceRect region;

    readonly double surface;

    #endregion

    #region Private / Constructors

    internal ModelScanner_Operation_Job(CubemapFaceRect region, IHeightmap heightmap,
      IGeometry geometry, Vec3D center, VerticalRange range, double surface,
      IModelCollider collider)
    {
      this.region = region;
      this.heightmap = heightmap;
      this.geometry = geometry;
      this.center = center;
      this.range = range;
      this.surface = surface;
      this.collider = collider;

      elevation = null;
      displacement = false;
    }

    internal ModelScanner_Operation_Job(CubemapFaceRect region, IHeightmap elevation,
      IHeightmap heightmap, IGeometry geometry, Vec3D center, VerticalRange range,
      IModelCollider collider)
    {
      this.region = region;
      this.elevation = elevation;
      this.heightmap = heightmap;
      this.geometry = geometry;
      this.center = center;
      this.range = range;
      this.collider = collider;

      displacement = true;
      surface = 0;
    }

    #endregion

    #region Private / Methods

    internal void Perform()
    {
      HeightmapRegion samples;
      int n;

      n = geometry.Map.Size;

      if (displacement)
      {
        CubemapFaceRect regionPad;
        NormalBuffer normals;

        regionPad = new CubemapFaceRect(region.X1 - 1,
          region.Y1 - 1,
          region.Width + 2,
          region.Height + 2,
          region.Face);

        samples = HeightmapRegion.Create(regionPad.Width,
          regionPad.Height,
          HeightmapLayer.Elevation | HeightmapLayer.Displacement | HeightmapLayer.Coverage);

        samples.Vertical = heightmap.Vertical;

        elevation.GetRegion(regionPad,
          HeightmapLayer.Coverage | HeightmapLayer.Elevation,
          1,
          samples);

        normals = new NormalBuffer(samples.Width, samples.Height);
        normals.ComputeNormals(geometry, null, samples, regionPad.ToFace(0, 0), 1);

        for (int y = 0; y < region.Height; y++)
        {
          for (int x = 0; x < region.Width; x++)
          {
            HeightmapSample sample;
            int normalOfs;
            Vec3I cube;
            Vec3D point;
            Vec3D position;
            Vec3D normal;
            Ray ray;
            double d1;
            double d2;

            cube = region.ToCube(x, y, n);

            sample = samples.GetAt(x + 1, y + 1);

            position = center.Add(
              geometry.ComputePosition(cube, range.DenormalizeElevation(sample.Elevation)));

            normalOfs = normals.OffsetAt(x + 1, y + 1);
            normal = new Vec3D(normals.Samples[normalOfs],
              normals.Samples[normalOfs + 1],
              normals.Samples[normalOfs + 2]);

            // Pick outwards (positive displacement).

            ray = new Ray(position, normal);
            d1 = collider.PickRay(ray, out point, out Vec3D _)
              ? ray.Distance(point)
              : Maths.MaxDouble;

            // Pick inwards (negative displacement).

            ray = new Ray(position, normal.Neg());
            d2 = collider.PickRay(ray, out point, out Vec3D _)
              ? -ray.Distance(point)
              : Maths.MaxDouble;

            // Use smaller absolute displacement.

            if (Maths.Abs(d1) > Maths.Abs(d2))
              d1 = d2;

            // No displacement if neither ray has hit.

            if (d1 == Maths.MaxDouble)
              d1 = 0;

            // Limit maximum ray distance with omni-directional distance.

            d1 *= Maths.Sqr(Maths.Min(1,
              Maths.Sqrt2 * collider.DistanceTo(position) / Maths.Abs(d1)));

            samples.SetAt(x + 1, y + 1, sample.SetDisplacement(range.NormalizeElevationClamp(d1)));
          }
        }

        heightmap.UpdateRegion(region, samples, false, HeightmapLayer.Displacement, 1, 1);
      }
      else
      {
        samples = HeightmapRegion.Create(region.Width,
          region.Height,
          HeightmapLayer.Elevation | HeightmapLayer.Coverage);

        samples.Clear(true, samples.Width, samples.Height);

        for (int y = 0; y < samples.Height; y++)
        {
          for (int x = 0; x < samples.Width; x++)
          {
            Vec3I cube;
            Vec3D up;
            Vec3D point;
            int elevation;

            cube = region.ToCube(x, y, n);

            up = geometry.ComputeUp(cube);

            if (!collider.PickRay(
              new Ray(center.Add(up.Mul((surface + range.Maximum) * 1.1)), up.Neg()),
              out point,
              out Vec3D _))
              continue;

            elevation = range.NormalizeElevationClamp(point.Sub(center).Length - surface);

            samples.SetAt(x, y, new HeightmapSample(elevation));
          }
        }

        heightmap.UpdateRegion(region, samples);
      }

      samples.Delete();
    }

    #endregion
  }
}