﻿// -------------------------------------------------------------------------------------------------
//          _____          _______                                   ____   _____
//      ,-:` \;',`'-,     |__  _(_)                                 |___ \ |  __ \
//    .'-;_,;  ':-;_,'.      | | _  _ __   _ __ ___    __ _  _ __     __) || |  | |
//   /;   '/    ,  _`.-\     | || || '_ \ | '_ ` _ \  / _` || '_ \   |__ < | |  | |
//  | '`. (`     /` ` \`|    | || || | | || | | | | || (_| || | | |  ___) || |__| |
//  |:.  `\`-.   \_   / |    |_||_||_| |_||_| |_| |_| \__,_||_| |_| |____/ |_____/
//  |     (   `,  .`\ ;'|
//   \     | .'     `-'/      Realtime terrain processing, rendering and analysis
//    `.   ;/        .'
//      `'-._____.-'`             http://www.tinman3d.com - info@tinman3d.com
// -------------------------------------------------------------------------------------------------
// FILE:       TerrainSpace.cs
// -------------------------------------------------------------------------------------------------
// COPYRIGHT:  Matthias Englert, me@tinman3d.com
// -------------------------------------------------------------------------------------------------
// AUTHOR(s):  Matthias Englert, me@tinman3d.com
// -------------------------------------------------------------------------------------------------
// DISCLAIMER: This source  code is  provided "as  is" and  without warranties  as to performance or
//             merchantability. The  author and/or  distributors of  this source  code may have made
//             statements about this source code.  Any such statements do not  constitute warranties
//             and shall not be  relied on by  the user in deciding whether to use this source code.
//             This source code is provided without any expressed or implied warranties  whatsoever.
//             Because of the diversity of conditions and hardware under which this source code  may
//             be used,  no warranty  of fitness  for a  particular purpose  is offered. The user is
//             advised to test the source code thoroughly before relying on it. The user must assume
//             the entire risk of using the source code.
// -------------------------------------------------------------------------------------------------

using CodeX;
using Tinman.Core;
using Tinman.Core.Math;
using Tinman.Core.Math.Vectors;
using Tinman.Engine.Components;
using Tinman.Engine.Rendering.Effects.Shared;
using Tinman.Terrain.Geometries;
using Tinman.Terrain.Georef;
using Tinman.Terrain.Kernel;

namespace Tinman.Demo.SIGGRAPH21
{
  /// <summary>
  ///   Provides coordinate transformation methods between terrain-space, light-space and
  ///   material-space.
  /// </summary>
  /// <remarks>
  ///   <para>
  ///     Terrain-space is the coordinate system that is established by <see cref="IGeometry"/>.
  ///     <br/>
  ///     See <see cref="TerrainMesh.Mesh"/> and <see cref="IMesh.Geometry"/>.
  ///   </para>
  ///   <para>
  ///     Light-space is the coordinate system used to perform lighting calculations, e.g. light
  ///     directions vectors and normal vectors are transformed into this coordinate system. Also,
  ///     <see cref="IEnvironmentMap"/> backgrounds are applied in light-space. <br/>
  ///     See <see cref="TerrainView.LightingNorth"/>.
  ///   </para>
  ///   <para>
  ///     Material-space is the coordinate system that is used to perform tri-planar material-based
  ///     texturing; it is similar to light-space, but not identical: in order to avoid texture
  ///     swimming, material-space must be translated and/or rotated. Also, a non-linear projection
  ///     must be applied for small planetary terrains. <br/>
  ///     See <see cref="TerrainLayerMaterial.Chunks"/>.
  ///   </para>
  ///   <para>
  ///     The steps for transforming light-space into material-space are described below.
  ///   </para>
  ///   <para>
  ///     The initial material coordinates <c>m0</c> are computed as follows:
  ///     <code>
  ///       // Transform to light-space, centered at material origin.
  ///       m0 = lightSpace * (point - origin)
  ///     </code>
  ///     where <c>lightSpace</c> is the inverse of <see cref="LightToTerrainMatrix"/>,
  ///     <c>origin</c> is <see cref="Origin"/> and <c>point</c> is the terrain-space point for
  ///     which material coordinates need to be computed. <br/>
  ///     This step establishes the orientation of material-space, with respect to horizontal and
  ///     vertical textures for tri-planar mapping. Also, floating-point precision issues are fixed
  ///     for large planetary terrains.
  ///   </para>
  ///   <para>
  ///     The projected material coordinates <c>m1</c> are computed as follows: <br/>
  ///     If <see cref="Projection"/> is <c>true</c> and the terrain geometry is geocentric (see
  ///     <see cref="IGeometry.ToGeocentric"/>):
  ///     <code>
  ///       // Compute coefficients.
  ///       horizontal = length(m0.X, m0.Z)
  ///       vertical   = radius + m0.Y
  ///       angle      = atan2(horizontal, vertical)
  ///       scale      = horizontal > 0 ? angle * radius / horizontal : 0
  ///       // Use height above geocentric sphere as vertical coordinate.
  ///       m1.y     = length(m0.X, vertical, m0.Z) - radius
  ///       // Scale horizontal vector to length of corresponding arc on sphere.
  ///       m1.xz    = m0.xz * scale;
  ///     </code>
  ///     where <c>offset</c> is <see cref="Offset"/> and <c>radius</c> is <see cref="Radius"/>.
  ///     <br/>
  ///     Otherwise, no projection is performed and the initial material coordinates are used:
  ///     <code>
  ///       m1 = m0
  ///     </code>
  ///     This step ensures that material-space is invariant to camera movement and rotation. For
  ///     large planetary terrains, material coordinate projection is not necessary, since the
  ///     visual error resulting from using unprojected material coordinates is not apparent.
  ///   </para>
  ///   <para>
  ///     The adjusted material coordinates <c>m2</c> are computed as follows:
  ///     <code>
  ///       // Get sine and cosine of rotation angle.
  ///       s    = sin(-rotation)
  ///       c    = cos(-rotation)
  ///       // Rotate around Y-axis.
  ///       m2.x = c * m1.X + s * m1.Z + offset.X
  ///       m2.y = m1.y + offset.Y
  ///       m2.z = -s * m1.X + c * m1.Z + offset.Z
  ///     </code>
  ///     where <c>rotation</c> is <see cref="Rotation"/>. <br/>
  ///     This step ensures that material coordinates stay in place, when the coordinate frame is
  ///     moving. Not doing so would result in texture swimming artefacts, which visually
  ///     intolerable.
  ///   </para>
  /// </remarks>
  [UsedImplicitly]
  public sealed class TerrainSpace
  {
    #region Public / Attributes

    /// <summary>
    ///   Allow <see cref="Move"/> to modify the material coordinate space?
    /// </summary>
    /// <remarks>
    ///   Defaults to <c>true</c>.
    /// </remarks>
    /// <value>
    ///   <c>true</c> to enable material coordinate space updates via <see cref="Move"/>, <br/>
    ///   <c>false</c> to disable.
    /// </value>
    public bool AllowMove
    {
      get { return allowMove; }
      set { allowMove = value; }
    }

    /// <summary>
    ///   Returns the rotation matrix from light-space to terrain-space.
    /// </summary>
    /// <value>The rotation matrix.</value>
    public Mat3D LightToTerrainMatrix
    {
      get { return lightToTerrain; }
    }

    /// <summary>
    ///   The north direction.
    /// </summary>
    /// <value>The unit-length north direction vector.</value>
    public Vec3D North
    {
      get { return north; }
    }

    /// <summary>
    ///   The material coordinate offset.
    /// </summary>
    /// <value>The offset vector.</value>
    public Vec3D Offset
    {
      get { return offset; }
      set { offset = value; }
    }

    /// <summary>
    ///   The origin of light-space, in terrain-space.
    /// </summary>
    /// <value>The light-space origin.</value>
    public Vec3D Origin
    {
      get { return origin; }
    }

    /// <summary>
    ///   Enable projection of material-space for geocentric terrains?
    /// </summary>
    /// <remarks>
    ///   <para>
    ///     The Y-coordinate of projected material-space is the height above the geocentric sphere.
    ///     The X- and Y-coordinates are obtained by applying a map projection (a variant of
    ///     <see cref="CoordinateOperationMethod.PolarStereographic"/>).
    ///   </para>
    ///   <para>Defaults to <c>false</c>.</para>
    /// </remarks>
    /// <value>
    ///   <c>true</c> to use projected material-space for texturing of geocentric terrains, <br/>
    ///   <c>false</c> to use unprojected material-space.
    /// </value>
    public bool Projection
    {
      get { return projection; }
      set { projection = value; }
    }

    /// <summary>
    ///   The geocentric radius.
    /// </summary>
    /// <value>The radius, in terrain-space.</value>
    public double Radius
    {
      get { return radius; }
    }

    /// <summary>
    ///   The material rotation angle.
    /// </summary>
    /// <value>The rotation angle, in radians.</value>
    public double Rotation
    {
      get { return rotation; }
      set { rotation = value; }
    }

    #endregion

    #region Public / Methods

    /// <summary>
    ///   Transforms a point from light-space to terrain-space.
    /// </summary>
    /// <param name="point">The point to transform, in light-space.</param>
    /// <returns>The transformed point, in terrain-space.</returns>
    [Pure]
    public Vec3D LightToTerrain(Vec3D point)
    {
      return lightToTerrain.Mul3(point).Add(origin);
    }

    /// <summary>
    ///   Transforms a point from material-space to terrain-space.
    /// </summary>
    /// <param name="point">The point to transform, in material-space.</param>
    /// <returns>The transformed point, in terrain-space.</returns>
    [Pure]
    public Vec3D MaterialToTerrain(Vec3D point)
    {
      return MaterialToTerrain(point, origin);
    }

    /// <summary>
    ///   Updates the terrain-space transformations.
    /// </summary>
    /// <remarks>
    ///   Use this method when moving over the terrain. This will keep material-space in place,
    ///   avoiding texture swimming artefacts. If <see cref="AllowMove"/> returns <c>true</c>, this
    ///   method will do nothing and will return silently.
    /// </remarks>
    /// <param name="origin">
    ///   Origin of light-space, in terrain-space. The origin will be moved along the terrain
    ///   up-vector to elevation zero.
    /// </param>
    /// <param name="north">The north-direction.</param>
    /// <param name="chunks">The material chunk factor.</param>
    public void Move(Vec3D origin, Vec3D north, double chunks)
    {
      Vec3D originBefore;
      Vec3D side;
      Vec3D a;
      Vec3D b;

      if (!allowMove)
        return;

      originBefore = this.origin;

      side = ComputeOrigin(origin, north);

      a = TerrainToMaterial(this.origin, originBefore);
      b = TerrainToMaterial(this.origin.Add(side), originBefore).Sub(a).Norm();

      ComputeMatrices(side);

      rotation = b.IsUndefined ? 0 : Maths.Atan2(b.Z, b.X);
      offset = a.Sub(offset.Div(chunks).Add(0.5, 0.5, 0.5).Floor().Mul(chunks));
    }

    /// <summary>
    ///   Resets the accumulated adjustment of material-space.
    /// </summary>
    /// <remarks>
    ///   Material-space is being adjusted in order to keep material textures in place. Otherwise,
    ///   texture swimming artefacts would occur. As a side-effect, this adjustment may result in
    ///   different material-spaces being used for the same location in terrain-space, when visited
    ///   at different points in time.
    /// </remarks>
    public void Reset()
    {
      offset = Vec3D.Zero;
      rotation = 0;
    }

    /// <summary>
    ///   Sets the terrain-space transformations.
    /// </summary>
    /// <remarks>
    ///   Use this method to restore all transformations to a previous state, for example after
    ///   loading a scene. This will make sure that material-based texturing will look exactly the
    ///   same.
    /// </remarks>
    /// <param name="origin">
    ///   Origin of light-space, in terrain-space. The origin will be moved along the terrain
    ///   up-vector to elevation zero.
    /// </param>
    /// <param name="north">The north-direction.</param>
    /// <param name="offset">The offset vector, see <see cref="Offset"/>.</param>
    /// <param name="rotation">The rotation angle, see <see cref="Rotation"/>.</param>
    public void Set(Vec3D origin, Vec3D north, Vec3D offset, double rotation)
    {
      ComputeMatrices(ComputeOrigin(origin, north));

      this.rotation = rotation;
      this.offset = offset;
    }

    /// <summary>
    ///   Transforms a point from terrain-space to light-space.
    /// </summary>
    /// <param name="point">The point to transform, in terrain-space.</param>
    /// <returns>The transformed point, in light-space.</returns>
    [Pure]
    public Vec3D TerrainToLight(Vec3D point)
    {
      return terrainToLight.Mul3(point.Sub(origin));
    }

    /// <summary>
    ///   Transforms a point from terrain-space to material-space.
    /// </summary>
    /// <param name="point">The point to transform, in terrain-space.</param>
    /// <returns>The transformed point, in material-space.</returns>
    [Pure]
    public Vec3D TerrainToMaterial(Vec3D point)
    {
      return TerrainToMaterial(point, origin);
    }

    /// <summary>
    ///   Updates the given render effect parameters.
    /// </summary>
    /// <param name="parameters">[not-null] The render effect parameters.</param>
    public void UpdateParameters(TerrainMaterialParameters parameters)
    {
#if DEBUG
      if (parameters == null)
        throw InvalidArgumentException.NotNull("TerrainSpace.UpdateParameters", "parameters");
#endif

      parameters.Origin = origin;
      parameters.Offset = Vec.To3F(offset);
      parameters.Rotation = rotation;
      parameters.Radius = IsProjection ? (float) radius : 0;
    }

    #endregion

    #region Private / Attributes

    bool allowMove;

    /// <summary>
    ///   The terrain geometry.
    /// </summary>
    readonly IGeometry geometry;

    bool IsProjection
    {
      get { return projection && geometry.ToGeocentric != null; }
    }

    /// <summary>
    ///   Light-to-terrain transformation matrix.
    /// </summary>
    Mat3D lightToTerrain;

    /// <summary>
    ///   The north direction.
    /// </summary>
    Vec3D north;

    /// <summary>
    ///   Offset in material-space.
    /// </summary>
    Vec3D offset;

    /// <summary>
    ///   Origin of light-space, in terrain-space.
    /// </summary>
    Vec3D origin;

    /// <summary>
    ///   Using projected material-space?
    /// </summary>
    bool projection;

    /// <summary>
    ///   The geocentric radius, in terrain-space.
    /// </summary>
    double radius;

    /// <summary>
    ///   Rotation in material-space.
    /// </summary>
    double rotation;

    /// <summary>
    ///   Terrain-to-light transformation matrix.
    /// </summary>
    Mat3D terrainToLight;

    /// <summary>
    ///   Up-vector of light-space, in terrain-space.
    /// </summary>
    Vec3D up;

    #endregion

    #region Private / Constructors

    internal TerrainSpace(IGeometry geometry)
    {
      this.geometry = geometry;

      lightToTerrain = Mat3D.Zero;
      offset = Vec3D.Zero;
      origin = Vec3D.Zero;
      terrainToLight = Mat3D.Zero;
      up = Vec3D.Zero;
      north = Vec3D.Zero;
      allowMove = true;
    }

    #endregion

    #region Private / Methods

    void ComputeMatrices(Vec3D side)
    {
      lightToTerrain = Mat3D.TangentSpace(side, up, side.Cross(up));
      terrainToLight = lightToTerrain.Transpose;
    }

    Vec3D ComputeOrigin(Vec3D origin, Vec3D north)
    {
      Vec3D coordinates;

      this.north = north.Norm();

      geometry.ComputeCoordinates(origin, out coordinates);
      geometry.ComputePosition(coordinates, 0, out this.origin, out up);

      radius = this.origin.Length;

      return this.north.Cross(up);
    }

    [Pure]
    Vec3D MaterialToTerrain(Vec3D point, Vec3D origin)
    {
      Vec3D result;

      result = lightToTerrain.Mul3(point).Add(origin);

      if (IsProjection)
      {
        for (int i = 0; i < 12; i++)
          result = result.Add(lightToTerrain.Mul3(point.Sub(TerrainToMaterial(result, origin))));
      }

      return result;
    }

    [Pure]
    Vec3D TerrainToMaterial(Vec3D point, Vec3D origin)
    {
      Vec3D material;
      double cos;
      double sin;

      // Compute initial material coordinates.

      material = terrainToLight.Mul3(point.Sub(origin));

      if (IsProjection)
      {
        double horizontal;
        double vertical;
        double scale;
        double angle;

        // Compute projected material coordinates.

        horizontal = Geometric.Length(material.X, material.Z);
        vertical = radius + material.Y;

        angle = Maths.Atan2(horizontal, vertical);

        scale = horizontal > 1E-10 ? angle * radius / horizontal : 0;

        material = new Vec3D(material.X * scale,
          Geometric.Length(material.X, vertical, material.Z) - radius,
          material.Z * scale);
      }

      // Compute adjusted material coordinates.

      sin = Maths.Sin(-rotation);
      cos = Maths.Cos(-rotation);

      return new Vec3D(cos * material.X + sin * material.Z,
        material.Y,
        -sin * material.X + cos * material.Z).Add(offset);
    }

    #endregion
  }
}