// -------------------------------------------------------------------------------------------------
//          _____          _______                                   ____   _____
//      ,-:` \;',`'-,     |__  _(_)                                 |___ \ |  __ \
//    .'-;_,;  ':-;_,'.      | | _  _ __   _ __ ___    __ _  _ __     __) || |  | |
//   /;   '/    ,  _`.-\     | || || '_ \ | '_ ` _ \  / _` || '_ \   |__ < | |  | |
//  | '`. (`     /` ` \`|    | || || | | || | | | | || (_| || | | |  ___) || |__| |
//  |:.  `\`-.   \_   / |    |_||_||_| |_||_| |_| |_| \__,_||_| |_| |____/ |_____/
//  |     (   `,  .`\ ;'|
//   \     | .'     `-'/      Realtime terrain processing, rendering and analysis
//    `.   ;/        .'
//      `'-._____.-'`             http://www.tinman3d.com - info@tinman3d.com
// -------------------------------------------------------------------------------------------------
// FILE:       Lighting.hlsl
// -------------------------------------------------------------------------------------------------
// COPYRIGHT:  Matthias Englert, me@tinman3d.com
// -------------------------------------------------------------------------------------------------
// AUTHOR(s):  Matthias Englert, me@tinman3d.com
// -------------------------------------------------------------------------------------------------
// DISCLAIMER: This source  code is  provided "as  is" and  without warranties  as to performance or
//             merchantability. The  author and/or  distributors of  this source  code may have made
//             statements about this source code.  Any such statements do not  constitute warranties
//             and shall not be  relied on by  the user in deciding whether to use this source code.
//             This source code is provided without any expressed or implied warranties  whatsoever.
//             Because of the diversity of conditions and hardware under which this source code  may
//             be used,  no warranty  of fitness  for a  particular purpose  is offered. The user is
//             advised to test the source code thoroughly before relying on it. The user must assume
//             the entire risk of using the source code.
// -------------------------------------------------------------------------------------------------

#ifndef TINMAN__LIGHTING
#define TINMAN__LIGHTING

#include "../GpuX.hlsl"
#include "CB_Lighting_Static.hlsl"
#include "Configuration.hlsl"
#include "SR_Lighting.hlsl"

// -------------------------------------------------------------------------------------------------
// Prototypes
// -------------------------------------------------------------------------------------------------

/*!
 * @brief Computes the ambient occlusion value.
 *
 * @param ao The input ambient occlusion value (vertex resp. occlusion map):
 *           0 := no ambient lighting
 *           1 := full ambient lighting
 *
 * @returns The output ambient occlusion value:
 *          0 := no ambient lighting
 *          1 := full ambient lighting
 */
float ComputeAmbientOcclusion(float ao);

/*!
 * @brief Computes a fake sample that approximates the material reflectance lookup texture.
 *
 * @param x The X-coordinate of the lookup vector.
 * @param y The Y-coordinate of the lookup vector.
 *
 * @returns The lookup sample.
 */
float2 ComputeFakeGgx(float x, float y);

/*!
 * @brief Computes the fog value.
 *
 * @param color    The base color.
 * @param eye      The unit-length eye vector.
 * @param distance The camera distance.
 *
 * @returns The fog color value.
 */
float3 ComputeFog(float3 color, float3 eye, float distance);

/*!
 * @brief Applies the tone mapping curve to the given HDR color.
 *
 * @param color The input HDR color.
 *
 * @returns The output LDR color.
 */
float3 ComputeToneMap(float3 color);

/*!
 * @brief Warps the Y-coordinate of the given cubemap direction vector.
 *
 * @param direction The direction vector.
 * @param warp      The spherical latitude offset, in radians.
 *
 * @returns The warped cubemap direction vector.
 */
float3 CubemapWarp(float3 direction, float warp);

/*!
 * @brief Samples the environment map with linear interpolation.
 *
 * Filtering will not be seamless for Direct3D 9.
 *
 * @param v   The cubemap texture coordinates.
 * @param lod The LOD-level to sample.
 *
 * @returns The sampled HDR color.
 */
float3 EnvironmentSample(float3 v, float lod);

/*!
 * @brief Samples the environment map with high-quality seamless filtering.
 *
 * @param v   The cubemap texture coordinates.
 * @param lod The LOD-level to sample.
 *
 * @returns The sampled HDR color.
 */
float3 EnvironmentSampleSeamless(float3 v, float lod);

/*!
 * @brief Performs bicubic interpolation on the environment cubemap texture.
 *
 * @param texCoords The cubemap texture coordinates, normalized to the XY plane.
 * @param m         The transformation from the normalized XY plane to cubemap coordinates.
 * @param lod       The discrete mipmap level to sample.
 *
 * @returns The interpolated cubemap texture sample.
 */
float3 TextureBicubic(float2 texCoords, float3x3 m, float lod);

/*!
 * @brief Computes the interpolation weights for a Uniform-B spline.
 *
 * @param v The saturated interpolation factor.
 *
 * @returns The weights of the four control points.
 */
float4 TextureBicubic_UniformB(float v);

/*!
 * @brief Wraps the given normalized cubemap texture coordinates.
 *
 * @param v The normalized cubemap texture coordinates in the XY plane.
 *
 * @returns The wrapped cubemap texture coordinates.
 */
float3 TextureBicubic_Wrap(float3 v);

// -------------------------------------------------------------------------------------------------
// Functions
// -------------------------------------------------------------------------------------------------

float ComputeAmbientOcclusion(float ao)
{
  return pow(max(0.001, ao), g_lighting.w);
}

float2 ComputeFakeGgx(float x, float y)
{
  float g;
  float r;

  // See Tinman.Engine.Models.ModelTextureTest.TestLookup.

  g = saturate(1.0 - sqrt(x * x * 12.0 + y * y * 3.5));
  x = 1.0 - x;
  y = 1.0 - y;
  r = saturate(0.31 + 0.69 * sqrt(x * x / 2.5 + y * y * 4.0) - g);

  return float2(r, g);
}

float3 ComputeFog(float3 color, float3 eye, float distance)
{
  float temp;
  float f;
  float3 fog;

  temp = distance / g_fogDistance;
  f = eye.y * eye.y;
  eye = float3(eye.x, min(0.0, eye.y), eye.z);
  fog = EnvironmentSample(-eye, g_mipmap.z);
  fog = lerp(fog, color, f);

  return lerp(fog, color, 1.0 / exp(temp * temp));
}

float3 ComputeToneMap(float3 color)
{
  // Perform no tone-mapping when low-dynamic range rendering is performed.

  if (g_environmentScale == 1.0)
    return color * g_lighting.z;

  // Apply tone-mapping when high-dynamic range rendering is performed.

  color = color * g_lighting.z / 0.6;
  color = mul(color, float3x3(0.59719, 0.076, 0.0284, 0.35458, 0.90834, 0.13383, 0.04823, 0.01566,
      0.83777));
  color = (color * (color + 0.0245786) - 9.0537E-5) / (color * (color * 0.983729 + 0.432951) +
      0.238081);
  color = mul(color, float3x3(1.60475, -0.10208, -0.00327, -0.53108, 1.10813, -0.07276, -0.07367,
      -0.00605, 1.07602));

  return clamp(color, 0.0, 1.0);
}

float3 CubemapWarp(float3 direction, float warp)
{
  direction = normalize(direction);

  if (warp > 0.0)
    direction.y = sin(asin(direction.y) + warp);

  return direction;
}

float3 EnvironmentSample(float3 v, float lod)
{
  return gxTextureCubeLod(g_environment, v, lod) * g_environmentScale;
}

float3 EnvironmentSampleSeamless(float3 v, float lod)
{
  if (CUBEMAP_SEAMLESS)
  {
    float3 a;
    float m;
    float3x3 t;
    float lod0;

    // Direct3D 10 and later support seamless cubemap filtering. The cubemap texture has been
    // generated without edge fixup. Bicubic interpolation is used for high-quality filtering of
    // the mipmap levels, which are then mixed linearly.

    a = abs(v);

    if (a.x > a.y && a.x > a.z)
    {
      m = v.x;
      t = float3x3(0, 1, 0, 0, 0, 1, 1, 0, 0);
    }
    else if (a.y > a.z)
    {
      m = v.y;
      t = float3x3(0, 0, 1, 1, 0, 0, 0, 1, 0);
    }
    else
    {
      m = v.z;
      t = float3x3(1, 0, 0, 0, 1, 0, 0, 0, 1);
    }

    // Transform to normalized XY plane.

    v = mul(t, v / m);
    t *= sign(m);
    lod0 = floor(lod);

    return lerp(TextureBicubic(v.xy, t, lod0), TextureBicubic(v.xy, t, lod0 + 1.0), lod - lod0);
  }
  else
  {
    float lod0;
    float3 a;
    float3 sample0;
    float3 sample1;

    // Direct3D 9 does not support seamless cubemap filtering. The cubemap texture has been
    // generated with edge fixup, which is applied here separately for each mipmap level. Then
    // trilinear interpolation is used.

    a = abs(v);

    if (a.x > a.y && a.x > a.z)
      a = float3(0, 1, 1);
    else if (a.y > a.z)
      a = float3(1, 0, 1);
    else
      a = float3(1, 1, 0);

    lod0 = floor(lod);
    a *= pow(2.0, lod0) * g_edgeFixup;
    sample0 = EnvironmentSample(v * (-a + 1.0), lod0);
    sample1 = EnvironmentSample(v * (a * -2.0 + 1.0), lod0 + 1.0);

    return lerp(sample0, sample1, lod - lod0);
  }
}

float3 TextureBicubic(float2 texCoords, float3x3 m, float lod)
{
  int size;
  float scale;
  float2 fraction;
  float2 texCoords0;
  float4 fx;
  float4 fy;
  float4 f;
  float4 offset;
  float3 s0;
  float3 s1;
  float3 s2;
  float3 s3;
  float3 result;

  // Setup 4x4 interpolation grid for Uniform-B spline.

  gxSizeOfLevel(g_environment, (int) lod, size);

  scale = 2.0 / size;
  fraction = texCoords / scale - 0.5;
  texCoords0 = floor(fraction);
  fraction -= texCoords0;

  // Convert the control point weights of the spline to linear mix factors.

  fx = TextureBicubic_UniformB(fraction.x);
  fy = TextureBicubic_UniformB(fraction.y);
  f = float4(fx.xz + fx.yw, fy.xz + fy.yw);
  offset = float4(texCoords0.x - 0.5 + fx.y / f.x, texCoords0.x + 1.5 + fx.w / f.y, texCoords0.y -
      0.5 + fy.y / f.z, texCoords0.y + 1.5 + fy.w / f.w) * scale;

  // Use the computed linear mix factors to fetch samples that are weighted according to the
  // weights of the spline control points.

  s0 = EnvironmentSample(mul(TextureBicubic_Wrap(float3(offset.xz, 1)), m), lod);
  s1 = EnvironmentSample(mul(TextureBicubic_Wrap(float3(offset.yz, 1)), m), lod);
  s2 = EnvironmentSample(mul(TextureBicubic_Wrap(float3(offset.xw, 1)), m), lod);
  s3 = EnvironmentSample(mul(TextureBicubic_Wrap(float3(offset.yw, 1)), m), lod);
  f.x /= f.x + f.y;
  f.z /= f.z + f.w;
  result = lerp(lerp(s3, s2, f.x), lerp(s1, s0, f.x), f.z);

  // To avoid sampling artefacts at the cubemap corners, create a smooth transition to linear
  // interpolation before artefacts start to occur, so that seamless cubemap filtering of the
  // GPU may take over.

  fraction = (-abs(texCoords) + 1.0) * size;
  scale = saturate((max(fraction.x, fraction.y) - 1.0) / 2.0);

  if (scale < 1.0)
  {
    result = lerp(EnvironmentSample(mul(float3(texCoords, 1), m), lod), result, scale);
  }

  return result;
}

float4 TextureBicubic_UniformB(float v)
{
  float3 f;

  f = float3(1, 2, 3) - v;
  f = f * f * f / 6.0;
  f.y -= 4.0 * f.x;
  f.z -= 10.0 * f.x + 4.0 * f.y;

  return float4(f, 1.0 - f.x - f.y - f.z);
}

float3 TextureBicubic_Wrap(float3 v)
{
  float2 a;
  float2 d;

  a = abs(v.xy);
  d = max(0.0, a - 1.0);

  if (d.x > d.y)
  {
    v.x = clamp(v.x, -1.0, 1.0);
    v.z -= d.x;
  }
  else if (d.y > 0.0)
  {
    v.y = clamp(v.y, -1.0, 1.0);
    v.z -= d.y;
  }

  return v;
}

#endif
