// -------------------------------------------------------------------------------------------------
//          _____          _______                                   ____   _____
//      ,-:` \;',`'-,     |__  _(_)                                 |___ \ |  __ \
//    .'-;_,;  ':-;_,'.      | | _  _ __   _ __ ___    __ _  _ __     __) || |  | |
//   /;   '/    ,  _`.-\     | || || '_ \ | '_ ` _ \  / _` || '_ \   |__ < | |  | |
//  | '`. (`     /` ` \`|    | || || | | || | | | | || (_| || | | |  ___) || |__| |
//  |:.  `\`-.   \_   / |    |_||_||_| |_||_| |_| |_| \__,_||_| |_| |____/ |_____/
//  |     (   `,  .`\ ;'|
//   \     | .'     `-'/      Realtime terrain processing, rendering and analysis
//    `.   ;/        .'
//      `'-._____.-'`             http://www.tinman3d.com - info@tinman3d.com
// -------------------------------------------------------------------------------------------------
// FILE:       Shadow.hlsl
// -------------------------------------------------------------------------------------------------
// COPYRIGHT:  Matthias Englert, me@tinman3d.com
// -------------------------------------------------------------------------------------------------
// AUTHOR(s):  Matthias Englert, me@tinman3d.com
// -------------------------------------------------------------------------------------------------
// DISCLAIMER: This source  code is  provided "as  is" and  without warranties  as to performance or
//             merchantability. The  author and/or  distributors of  this source  code may have made
//             statements about this source code.  Any such statements do not  constitute warranties
//             and shall not be  relied on by  the user in deciding whether to use this source code.
//             This source code is provided without any expressed or implied warranties  whatsoever.
//             Because of the diversity of conditions and hardware under which this source code  may
//             be used,  no warranty  of fitness  for a  particular purpose  is offered. The user is
//             advised to test the source code thoroughly before relying on it. The user must assume
//             the entire risk of using the source code.
// -------------------------------------------------------------------------------------------------

#ifndef TINMAN__SHADOW
#define TINMAN__SHADOW

#include "../GpuX.hlsl"
#include "CB_Shadow_Static.hlsl"
#include "PS_Out_Target.hlsl"
#include "SR_Shadow.hlsl"

// -------------------------------------------------------------------------------------------------
// Prototypes
// -------------------------------------------------------------------------------------------------

/*!
 * @brief Computes the shadow for the given point.
 *
 * @param position xyz = Point in world-space for which to compute a shadow map sample.
 *                 w = Z-coordinate of point, in camera-space.
 *
 * @returns The shadow attenuation factor (0 := full shadow, 1 := no shadow).
 */
float ComputeShadow(float4 position);

#ifdef GX_SHADER_PS
/*!
 * @brief Writes a shadow map sample for the given depth.
 *
 * @param depth The shadow depth in the range [0..1].
 * @param alpha The alpha value to write.
 *
 * @returns The depth values to write into the shadowmap:
 *          r = depth
 *          g = squared depth
 *          The alpha value to use for alpha-to-coverage:
 *          a = alpha
 */
PS_Out_Target WriteShadow(float depth, float alpha = 1.0);
#endif

/*!
 * @brief Computes the shadow for the given point.
 *
 * @param position     Point in world-space for which to compute a shadow map sample.
 * @param index        The shadow cascade index.
 * @param shadowMatrix The shadow cascade matrix.
 *
 * @returns The shadow attenuation factor (0 := full shadow, 1 := no shadow).
 */
float ComputeShadow(float3 position, int index, float4x4 shadowMatrix);

// -------------------------------------------------------------------------------------------------
// Functions
// -------------------------------------------------------------------------------------------------

float ComputeShadow(float4 position)
{
  float2 f;
  float s;

  f = saturate(g_shadowBlend.xz * position.w + g_shadowBlend.yw);

  if (f.x == 0.0)
  {
    // Only near cascade

    s = ComputeShadow(position.xyz, 0, g_shadowMatrix[0]);
  }
  else if (f.y == 1.0)
  {
    // Only far cascade

    s = ComputeShadow(position.xyz, 2, g_shadowMatrix[2]);
  }
  else
  {
    // Always mid cascade

    s = ComputeShadow(position.xyz, 1, g_shadowMatrix[1]);

    // Mix mid cascade with near or far?

    if (f.x < 1.0)
      s = lerp(ComputeShadow(position.xyz, 0, g_shadowMatrix[0]), s, f.x);
    else if (f.y > 0.0)
      s = lerp(s, ComputeShadow(position.xyz, 2, g_shadowMatrix[2]), f.y);
  }

  return s;
}

#ifdef GX_SHADER_PS
PS_Out_Target WriteShadow(float depth, float alpha)
{
  PS_Out_Target result;
  float dx;
  float dy;

  depth = saturate(depth);
  dx = ddx(depth);
  dy = ddy(depth);
  result.Color = float4(depth, depth * depth + (dx * dx + dy * dy) / 4.0, 0, alpha);

  return result;
}
#endif

float ComputeShadow(float3 position, int index, float4x4 shadowMatrix)
{
  float4 temp;
  float2 shadow;
  float variance;
  float value;

  // Transform to shadowmap-space and fetch shadowmap sample.

  temp = mul(shadowMatrix, float4(position, 1));
  temp.xyz /= temp.w;
  shadow = gxTextures2D(g_shadowMap, index, temp.xy);

  // Apply variance-based shadow mapping.

  variance = max(2.5E-5, shadow.y - shadow.x * shadow.x);
  value = shadow.x - temp.z;

  return max(smoothstep(g_lightBleed, 1.0, variance / (variance + value * value)), temp.z <= shadow.
      x ? 1.0 : 0.0);
}

#endif
